{{

Fun Game
---Robot Component


Stingray RC Control

Has two main components:
---Wireless driving via a range of control values
---RFID reading and wireless transmission

Uses the following hardware:
---XBEE
---Parallax RFID Reader
---Stingray robot
}}

CON

  _xinfreq = 5_000_000
  _clkmode = xtal1 + pll16x

  'Motor pin definitions
  lmforward = 25 
  lmbackward = 24
  rmforward = 26
  rmbackward = 27

  period = 5000 'for PWM object

  delta = 5 'the greatest change in speed per loop iteration

  RFID_EN = 16 'enable pin

  delta_x = 200
  delta_y = 20

  {{
  This program uses two methods to make the robot easier to control.

  First, the input is filtered through a piecewise linear function.
  This function gives the input less weight at either extreme, which
  means that at very slow speeds a larger joystick motion is required
  to change the same amount of speed as opposed to in the middle range,
  for example. This results in the user having more control at very
  slow and fast speeds, but not middle speeds.

  Second, the input is filtered through an artifical momentum algorithm.
  This accounts for the fact that the motors are very powerful, and
  respond almost instantly to input. While it may be good for a motor
  controller, for a human it is very hard to control such a fast vehicle.
  The artifical momentum helps to detune the maximum change of motor
  speed. The robot is given twice as much momentum speeding up as
  slowing down. The asymetry is due to testing with user feedback, which
  found it was easier to control if the robot accelerated less rapidly
  than it slowed down.
  }}


  {
  'Use the following formulas to calculate the piecewise function values
  'Define delta_x and delta_y first

  y_imx = (((1024 * delta_y) - 51200)/delta_x) + 50
  m_mx = (50-delta_y)/delta_x
  m_in = delta_y / delta_x
  y_0in = 0
  y_1in = 100 - (delta_y * 2048) / (1024 - delta_x)
}
  y_imx = -104
  m_mx_n = 3
  m_mx_d = 20
  m_in_n = delta_y
  m_in_d = 1024-delta_x
  y_0in = 0
  y_1in = 50

OBJ
  com  : "FullDuplexSerialPlus"
  pwm  :  "PWM_32_v2"
  rfid : "FullDuplexSerialPlus"
  'debug: "FullDuplexSerialPlus"
  
VAR
  long rfidstack[20]
  byte tag[11]
  byte run_rfid 'the button value is stored here

PUB Main | input, lspeed, rspeed, prev_lspeed, prev_rspeed, lsign, rsign, temp, i

  pwm.start
  com.start(15, 14, 0, 57600)
  rfid.start(17, 23, 0, 2400) '23 is a dud pin
  'debug.start(31, 30, 0, 115200)

  lspeed := 0
  rspeed := 0

  run_rfid := 0

  waitcnt(clkfreq/4 + cnt)
  cognew(read_rfid, @rfidstack)
  waitcnt(clkfreq/4 + cnt)
  'debug.str(string("mobile robot..."))

  repeat
    'The characters y and z are for synchronization
    repeat until com.rx == "y"
    temp := com.rx
    run_rfid := temp & %1 'extract the low bit

    repeat until com.rx == "z"
    lspeed := 2048 - (((com.rx << 8) + com.rx)) +44
    rspeed := 2048 - (((com.rx << 8) + com.rx)) -28

      'filter out noise when at rest
    if(|| lspeed < 10)
      lspeed := 0

    if(|| rspeed < 10)
      rspeed := 0

{
    debug.str(string("Speeds, l r: "))
    debug.dec(lspeed)
    debug.str(string("     "))
    debug.dec(rspeed)
    debug.tx(13)
}
    'lspeed := (^^lspeed)*3*lsign
    'rspeed := (^^rspeed)*3*rsign

    lspeed := filter_input(lspeed, @prev_lspeed)
    rspeed := filter_input(rspeed, @prev_rspeed)

{

    debug.str(string("Speeds, l r:                     "))
    debug.dec(lspeed)
    debug.str(string("     "))
    debug.dec(rspeed)
    debug.tx(13)
}

    if(rspeed => 0)
      pwm.duty(rmbackward, 0,0)
      pwm.duty(rmforward, rspeed, period)
    else
      pwm.duty(rmbackward, ||rspeed, period)
      pwm.duty(rmforward, 0, 0)

    if(lspeed => 0)
      pwm.duty(lmbackward, 0,0)
      pwm.duty(lmforward, lspeed, period)

    else
      pwm.duty(lmbackward, ||lspeed, period)
      pwm.duty(lmforward, 0, 0)

PUB filter_input(speed, prev_speed) : final_speed | sign, prev_sign
'prev_speed is a long address

  sign := 1
  if speed < 0
    sign := -1
    ||speed

  prev_sign := 1
  if LONG[prev_speed] < 0
    prev_sign := -1
    || LONG[prev_speed]

  if(speed < (1024 - delta_x))
    'LHS region
    'debug.tx("L")
    speed := ((m_in_n * speed)/m_in_d) + y_0in
  elseif(speed > (1024 + delta_x))
    'RHS region
    'debug.tx("R")
    speed := ((m_in_n * speed)/m_in_d) + y_1in
  else
    'CENTRAL region
    'debug.tx("C")
    speed := ((m_mx_n * speed)/m_mx_d) + y_imx


  'Add in artificial Momentum
  if sign == 1 AND prev_sign == 1
    if speed > (LONG[prev_speed] + delta) 'faster
      speed := (LONG[prev_speed] + delta)
    elseif speed < (LONG[prev_speed] - (delta*2)) 'slower
      speed := (LONG[prev_speed] - (2*delta))

  elseif sign == -1 AND prev_sign == -1
    if speed > (LONG[prev_speed] + delta) 'faster
      speed := (LONG[prev_speed] + delta)
    elseif speed < (LONG[prev_speed] - (2*delta)) 'slower
      speed := (LONG[prev_speed] - (2*delta))
    speed := speed * sign

  elseif sign == 1 AND prev_sign == -1
    speed := (|| LONG[prev_speed] - delta) * prev_sign
  else 'sign == -1 AND prev_sign == 1
    speed := (|| LONG[prev_speed] - delta) * prev_sign


  final_speed := speed '* sign
  LONG[prev_speed] := final_speed

  final_speed := -98 #> final_speed <# 98

  'stop annoying buzzing when sitting still
  if ||final_speed < 2
    final_speed := 0

  return final_speed


PUB read_rfid | temp, i, found, time, should_send
{{
The RFID reading has two main issues:
---The card must be precisely underneath the robot
---The card must be in position for a (relatively) long time

This is due to the slowness of the RFID reader (2400
baud!) and the limited field of the reader.

The solution is just to stop on top of the card.

}}


  rfid.rxflush
  'enable RFID
  DIRA[RFID_EN]~~
  OUTA[RFID_EN]~

  should_send := true
  repeat
    temp := rfid.rxcheck

    'found card, now read in the ID
    if(temp == $0A)
      'we get here we know to expect an rfid card
      'debug.str(string(13, "Found RFID "))
      repeat i from 0 to 9
        temp := rfid.rx
        'debug.tx(temp)
        tag[i] := temp
      tag[i] := 0 'note: i == 10 here
      'debug.tx(13)

    'should we output to wireless
    'debug.dec(run_rfid)
    'debug.tx(13)
    if run_rfid == 1 AND should_send
      com.tx($0A)
      com.str(@tag)
      should_send := false

    if run_rfid == 0
      should_send := true

    waitcnt(clkfreq / 250 + cnt) 'delay 4 ms

